﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;

namespace TestSqlBuilder
{
    public class SqlMapper
    {
        //提供一个Dictionary，以及类型映射
        protected string pSymbol = "@";//参数符号
        protected string cSymbol = "+";//连接符号 
        private static Regex reg = new Regex(@"^(?<tab>\w+\.)?(?<Pre>(PK|Begin|End|Like|UnLike|Null))?(?<Key>\w+)$");
        private StringBuilder strCondition = new StringBuilder();
        public string Condition
        {
            get
            {
                if (strCondition.Length > 0) strCondition.Length -= 3;
                return strCondition.ToString();
            }
        }
        private Dictionary<string, object> paras = new Dictionary<string, object>();//条件参数值
        public Dictionary<string, object> Paras { get { return paras; } }
        //类型缓存
        private static readonly ConcurrentDictionary<RuntimeTypeHandle, IEnumerable<PropertyInfo>> TypeProperties = new ConcurrentDictionary<RuntimeTypeHandle, IEnumerable<PropertyInfo>>();

        //单表查询
        public void SelectSingle<T>(Dictionary<string, object> entity, string tableName = "", bool keepParas = false)
        {
            if (!keepParas) Clear();
            if (tableName.Length > 0) tableName += ".";//如果表名
            var type = typeof(T);
            foreach (var single in entity)//遍历entity
            {
                var match = reg.Match(single.Key); //正则获取到属性对应的字段名
                var pre = match.Groups["Pre"].Value;//获取命名规则中的比较谓词
                var colName = match.Groups["Key"].Value;//获取字段名称
                var paraName = single.Key;//检测默认参数名是否被使用
                var seq = 0;//保存被字段被占用后默认分配的序号
                while (true)
                {
                    if (paras.ContainsKey(paraName))
                        paraName += (++seq).ToString();
                    else
                        break;
                }
                strCondition.AppendFormat(" {0} AND", Builder(pre, colName, paraName, tableName, single.Value));//拼接SQL条件
                //若该字段的类型为字符串，检测在Mapper中的类型
                if (single.Value.GetType() == typeof(string))
                {
                    var t = TypePropertiesCache(typeof(T), colName);
                    if (t == null)
                        throw new Exception(string.Format("在对象中没有找到给属性，对象:{0},属性:{1}", type, colName));
                    if (t != typeof(string))
                        paras[paraName] = Convert.ChangeType(single.Value, t);
                    else
                        paras[paraName] = single.Value;
                }
                else //若字段的类型非字符串，则无需检测和转换
                    paras[paraName] = single.Value;
            }
        }

        //联表查询
        public void SelectMulity<T, K>(Dictionary<string, Dictionary<string, object>> entity, string tableName1, string tableName2)
        {
            Clear();
            SelectSingle<T>(entity[tableName1], tableName1, true);
            SelectSingle<K>(entity[tableName2], tableName2, true);
        }

        //查询无需类型映射
        public void Select(Dictionary<string, object> entity)
        {
            Clear();
            foreach (var single in entity)//遍历entity
            {
                var match = reg.Match(single.Key); //正则获取到属性对应的字段名
                var pre = match.Groups["Pre"].Value;//获取命名规则中的比较谓词
                var colName = match.Groups["Key"].Value;//获取字段名称
                var table = match.Groups["tab"].Value;//获取字段表名限定
                var paraName = string.Format("{0}{1}", pre, colName);//获取该条件被分配的过程参数名
                var seq = 0;//保存被字段被占用后默认分配的序号
                while (true)
                {
                    if (paras.ContainsKey(paraName))
                        paraName += (++seq).ToString();
                    else
                    {
                        paras[paraName] = single.Value;
                        break;
                    }
                }
                strCondition.AppendFormat(" {0} AND",
                    Builder(pre, colName, paraName, table, single.Value));//拼接SQL条件
            }
        }

        #region 私有方法
        private string Builder(string pre, string colName, string paraName, string table, object colValue = null)
        {
            switch (pre)
            {
                case "Like": return Like(colName, paraName, table);
                case "UnLike": return Like(colName, paraName, table, false);
                case "Begin": return Begin(colName, paraName, table);
                case "End": return End(colName, paraName, table);
                case "Equal": return Equal(colName, paraName, table);
                case "Null":
                    bool value = false;
                    if (colValue != null && bool.TryParse(colValue.ToString(), out value) && value) ;
                    return Null(colName, table, value);
                default: return Equal(colName, paraName, table);
            }
        }

        private static Type TypePropertiesCache(Type type, string propertyName)
        {
            IEnumerable<PropertyInfo> pis;
            if (!TypeProperties.TryGetValue(type.TypeHandle, out pis))
            {
                pis = type.GetProperties().Where(p => p.CanWrite);
                TypeProperties[type.TypeHandle] = pis;
            }
            var t = pis.SingleOrDefault(p => p.Name == propertyName);
            if (t != null) return t.PropertyType;
            return null;
        }

        private void Clear()
        {
            strCondition.Length = 0;
            paras.Clear();
        }
        #endregion

        #region SQL拼接
        protected virtual string Equal(string colName, string paraName, string tab)
        {
            return string.Format("{0}{1} = {2}{3}", tab, colName, pSymbol, paraName);
        }
        protected virtual string Like(string colName, string paraName, string tab, bool not = true)
        {
            return string.Format("{0}{1} {2}like '%'{3}{4}{5}{3}'%'", tab, colName, not ? string.Empty : "not ", cSymbol, pSymbol, paraName);
        }
        protected virtual string Begin(string colName, string paraName, string tab)
        {
            return string.Format("{0}{1} >= {2}{3}", tab, colName, pSymbol, paraName);
        }
        protected virtual string End(string colName, string paraName, string tab)
        {
            return string.Format("{0}{1} <= {2}{3}", tab, colName, pSymbol, paraName);
        }
        protected virtual string Null(string colName, string tab, bool isNull)
        {
            return string.Format("{0}{1} is {2}null", tab, colName, isNull ? string.Empty : "not ");
        }
        #endregion
    }
}
